// Video Lines Effect Plugin. Created & Copyright (C) 1998 by Kevin Phillips
//
// Style : ImageFilter Special Effect.
//
// Creates a Video Scanline effect over the final render... Can specify the line
// thickness, how dark the scanlines are and add in antialiasing effects on the edges...

// First, Include all the necessary headers.
// The LW SDK headers first
#include <splug.h>                              // Server Plugin Header
#include <moni.h>                               // Plugin Progress Monitor Header
#include <lwran.h>                              // The Layout Library header
#include <lwpanel.h>                            // LW GUI Header

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

const char * fxtype[][7] = {"(None)","Above","Below","Both"};
int panlstate;
const char * const SPACE = " ";
int i;

// ... Then some system based headers if necessary
// #include <stdlib.h>
// #include <strstream.h>


// The Prototypes for the plugin.

                int                     activate(long, GlobalFunc *, void *, void *);
                LWInstance              create(LWError *);
                void                    destroy(LWInstance);
                LWError                 copy(LWInstance, LWInstance);
                LWError                 load(LWInstance, const LWLoadState *);
                LWError                 save(LWInstance, const LWSaveState *);
                void                    process(LWInstance, const FilterAccess *);
                unsigned int            flags(LWInstance);
                const char *            descln(LWInstance);
                int                     theinterface(long, GlobalFunc *, LWInstance, void *);



// The Instance for this plugin.

int spln,dirn,thickness;
double darken,brighten;

typedef struct VIDIMAGE_TAG{
    int spln;
    int dirn;
    int thickness;
    double darken;
    double brighten;
} VIDIMAGE;


// The Server record that tells LW what and where our plugin exists
// Note this appears in the LW Drop down list of plugin types...

ServerRecord ServerDesc[]=
       {
             {"ImageFilterHandler","T42_Vidlines", activate},
             {"ImageFilterInterface","T42_Vidlines", theinterface},
             {NULL}
       };


// These globals were created to make panel controls easier to add for your
// Interface.
static LWPanControlDesc desc;
static LWValue ival={LWT_INTEGER},ivecval={LWT_VINT},fval={LWT_FLOAT},
                        fvecval={LWT_VFLOAT},sval={LWT_STRING};



// Activation routine.

XCALL_(int)
activate(long version,GlobalFunc *global,ImageFilterHandler *local, void *data)
{
        // First, convert our instance to the right structure (Makes life easier!)

        // Designed for LW 5.0 and above...

        if (version !=2)        return AFUNC_BADVERSION;

                // OK, set up all the callback routines.  As you can see, by reinterpret
                // cast, the local pointer now contains a handler structure that can be
                // easily set up...

                local->create           = create;
                local->destroy          = destroy;
                local->copy             = copy;
                local->load             = load;
                local->save             = save;
                local->descln           = descln;
                local->process          = process;
                local->flags            = flags;

                // And upon setup, return an OK and ready to go value!

                return AFUNC_OK;
}


// Creation Function.

XCALL_(LWInstance) create(LWError *error)
{
                // Create an instance of the class 'VIDIMAGE' (ie. Where our data would
                // reside!)

                // VIDIMAGE *lwinst = new VIDIMAGE;
                VIDIMAGE *lwinst;
                XCALL_INIT;

        lwinst = malloc(sizeof(struct VIDIMAGE_TAG));
        if (!lwinst)
        {
                *error = "Error allocating memory!";
                return NULL;
        }

                // Place in some default values... These seem to work nicely for most images...

                lwinst->thickness = 1;
                lwinst->spln = 4;
                lwinst->darken = 0.75;
                lwinst->brighten = 0.0;
                lwinst->dirn = 3;

                // Now, convert this Instance to the LWInstance and return its pointer
                // That's all that's needed here!

                // return (reinterpret_cast<LWInstance>(lwinst));
                return((LWInstance)lwinst);

}



// Destroy Function

XCALL_(void) destroy(LWInstance *lwinst)
{
                // First, check that the instance exists, then destroy it!

                if (lwinst)
                    free(lwinst);
                return;
}


// Copy Function

XCALL_(LWError) copy(LWInstance from, LWInstance to)
{
                VIDIMAGE *first, *second;

                // Create the correct pointers using our friend reinterpret_cast<>
                // Convert the LWInstance's to the VIDIMAGE data structure...

                first = (VIDIMAGE *)(from);
                second = (VIDIMAGE *)(to);

                // Now copy the pointers so that the Data is ready to go!

                *second = *first;

                // And exit

                return NULL;
}




#define	BSWAP_L(l)
void writeInt(const LWSaveState	*sState, int dval)
{
	if (sState->ioMode == LWIO_OBJECT) {
		int		 val = dval;
		BSWAP_L(val);
		(*sState->write) (sState->writeData, (char *) &val, 4);
	} else {
		char		 buf[50];
		sprintf (buf, " %d", dval);
		(*sState->write) (sState->writeData, buf, strlen (buf) + 1);
	}
}

int readInt(const LWLoadState	*lState)
{
	int		 val=0;
	if (lState->ioMode == LWIO_OBJECT) {
		(*lState->read) (lState->readData, (char *) &val, 4);
		BSWAP_L(val);
		return val;
	} else {
		char		 buf[50];
		(*lState->read) (lState->readData, buf, 50);
		sscanf (buf, "%d", &val);
		return val;
	}
}

void writeDouble(const LWSaveState	*sState, double dval)
{
	if (sState->ioMode == LWIO_OBJECT) {
		double		 val = dval;
		BSWAP_L(val);
		(*sState->write) (sState->writeData, (char *) &val, sizeof(double));
	} else {
		char		 buf[50];
		sprintf (buf, " %f", dval);
		(*sState->write) (sState->writeData, buf, strlen (buf) + 1);
	}
}

double readDouble(const LWLoadState	*lState)
{
	if (lState->ioMode == LWIO_OBJECT) {
		double		 val;
		(*lState->read) (lState->readData, (char *) &val, sizeof(double));
		BSWAP_L(val);
		return val;
	} else {
		char		 buf[50];
		double		 val;
		(*lState->read) (lState->readData, buf, 50);
		sscanf (buf, "%lf", &val);
		return val;
	}
}






// Save Function

XCALL_(LWError) save(LWInstance inst, const LWSaveState *lwss)
{

                // Create an instance of FADEOUT
                VIDIMAGE *saveptr = (VIDIMAGE *)(inst);


                writeInt(lwss, saveptr->spln);
				writeInt(lwss, saveptr->dirn);
				writeInt(lwss, saveptr->thickness);
				writeDouble(lwss, saveptr->darken);
				writeDouble(lwss, saveptr->brighten);
/*
                // Create/allocate a character array to store the settings into.
                char saveline[64];

                // Define an output stream to write console out into our array
                // ostrstream ostr(saveline,64);

                // Fill the data with nulls...
                for (i=0; i<64; i++) saveline[i]='\0';

                // Write the data into the array
                // ostr << saveptr->spln << " " << saveptr->dirn << " " << saveptr->thickness << " ";
                // ostr << saveptr->darken << " " << saveptr->brighten;

               strcpy (saveline, (const char)saveptr->spln); strcpy (saveline, SPACE);
               strcpy (saveline, (const char)saveptr->dirn); strcpy (saveline, SPACE);
               strcpy (saveline, (const char)saveptr->thickness); strcpy (saveline, SPACE);
               strcpy (saveline, (const char)saveptr->darken); strcpy (saveline, SPACE);
               strcpy (saveline, (const char)saveptr->brighten);

                // Now write the data into the file...
                //lwss->write(lwss->writeData, saveline, 64);
                lwss->write(lwss->writeData, (char *)&saveline, sizeof(VIDIMAGE));
*/
                return (NULL);
}



// Load Function

XCALL_(LWError) load(LWInstance inst, const LWLoadState *lwls)
{
                // Create an instance of FADEOUT and reinterpret_cast<FADEOUT *>(inst);
                VIDIMAGE*loadptr = (VIDIMAGE *)(inst);


				loadptr->spln = readInt(lwls);
				loadptr->dirn = readInt(lwls);
				loadptr->thickness = readInt(lwls);
				loadptr->darken = readDouble(lwls);
				loadptr->brighten = readDouble(lwls);
								
                return (NULL);


/*  winuae crash if wrong values

                // Create/allocate a character array to store the settings into.
                char loadline[64];
                int ch;
                FILE *stream;

                // Read the file into the array
                lwls->read(lwls->readData, loadline, 64);

                stream = fopen (loadline, strlen (loadline));
                while ((ch = fgetc (stream)) != EOF)
                    strcpy (loadptr->spln, ch);
                    strcpy (loadptr->dirn, ch);
                    strcpy (loadptr->thickness, ch);
                    strcpy (loadptr->darken, ch);
                    strcpy (loadptr->brighten, ch);
                fclose (stream);

                // Define an input stream to read our array
                istrstream istr(loadline,64);

                // Now read the array back into our settings
                istr >> loadptr->spln;
                istr.ignore(1);         // Ignore the Spaces
                istr >> loadptr->dirn;
                istr.ignore(1);         // Ignore the Spaces
                istr >> loadptr->thickness;
                istr.ignore(1);         // Ignore the Spaces
                istr >> loadptr->darken;
                istr.ignore(1);         // Ignore the Spaces
                istr >> loadptr->brighten;

                return NULL;
*/
}



// Description Function - Appears in the Plugin button (When set)

XCALL_(const char*) descln(LWInstance inst)
{
                // Here it is!
                return "T42 Video-Lines";
}




// Flags

XCALL_(unsigned int) flags(LWInstance inst)
{
                return NULL;
}



// Support Function - Boosts an RGB value and passes back an integer

XCALL_(int) boostup (int color1, double bright)
{

        int finalval = color1 + (int)((double)color1 * bright);

        if (finalval < 0) finalval = 0;
        if (finalval > 255) finalval = 255;

        return finalval;
}


// Process Function

XCALL_(void) process(LWInstance inst, const FilterAccess *fa)
{
                // First, we would usually convert the Instance to get the parameters

                VIDIMAGE *procinst = (VIDIMAGE *)(inst);

                // Define any variables we need. In this case I'm going to play with
                // RGB values, so I need a small array and 4 pointers to the RGBA values
                // The ur and br's are for the antialiasing (Soft effect)

                BufferValue trgb[3], rgb[3], *r, *g, *b, *ur, *ug, *ub, *br, *bg, *bb;

                // And some variables for loops - In this case the height and width of
                // the image being filtered, and some loop variables
                unsigned short int i,j,k,l;


                // First Loop = From top line to the end of the image.
                // Note that the entire image must be processed otherwise you get
                // a nice black picture!.

                i = 0;

                while (i<fa->height)    {

                                // Now determine if the line is for changing...
                                // If so, we need to also work out soft effects if required and apply...
                                if (!(i % (procinst->spln))) {

                                        // Ok, we've determined that the line is to be drawn across...

                                        // First check for soft effect above/both
                                        if (procinst->dirn == 1 || procinst->dirn == 3) {

                                                                // Make sure we aren't at the top of the screen

                                                                if (i > 2) {

                                                                        // Get the RGB buffers of the pixels above
                                                                        ur = fa->bufLine(LWBUF_RED, i-2);
                                                                        ug = fa->bufLine(LWBUF_GREEN, i-2);
                                                                        ub = fa->bufLine(LWBUF_BLUE, i-2);

                                                                        br = fa->bufLine(LWBUF_RED, i-1);
                                                                        bg = fa->bufLine(LWBUF_GREEN, i-1);
                                                                        bb = fa->bufLine(LWBUF_BLUE, i-1);

                                                                        // Antialias the line
                                                                        for (l = 0; l<fa->width; l++) {
                                                                                trgb[0] = boostup(((ur[l] + br[l] + (int)((double)r[l] * procinst->darken))/3), procinst->brighten);
                                                                                trgb[1] = boostup(((ug[l] + bg[l] + (int)((double)g[l] * procinst->darken))/3), procinst->brighten);
                                                                                trgb[2] = boostup(((ub[l] + bb[l] + (int)((double)b[l] * procinst->darken))/3), procinst->brighten);

                                                                                // And draw the new line, now antialiased!
                                                                                fa->setRGB(l,i-1,trgb);
                                                                        }

                                                                }

                                                        } // End of the Above check...


                                                        // Now check for below/both
                                                        // Note that strangely, we can't read forward, so Below is done
                                                        // for the above line instead...
                                                        if (procinst->dirn == 2 || procinst->dirn == 3) {

                                                                // Make sure we aren't at the top (ie. as doing line above!)

                                                                if (i >= (procinst->spln - procinst->thickness)) {

                                                                        // Get the RGB buffers of the pixels below
                                                                        ur = fa->bufLine(LWBUF_RED, (i-procinst->spln)+procinst->thickness+1);
                                                                        ug = fa->bufLine(LWBUF_GREEN, (i-procinst->spln)+procinst->thickness+1);
                                                                        ub = fa->bufLine(LWBUF_BLUE, (i-procinst->spln)+procinst->thickness+1);

                                                                        br = fa->bufLine(LWBUF_RED, (i-procinst->spln)+procinst->thickness+2);
                                                                        bg = fa->bufLine(LWBUF_GREEN, (i-procinst->spln)+procinst->thickness+2);
                                                                        bb = fa->bufLine(LWBUF_BLUE, (i-procinst->spln)+procinst->thickness+2);

                                                                        // Antialias the line
                                                                        for (l = 0; l<fa->width; l++) {
                                                                                trgb[0] = boostup(((ur[l] + br[l] + (int)((double)r[l] * procinst->darken))/3), procinst->brighten);
                                                                                trgb[1] = boostup(((ug[l] + bg[l] + (int)((double)g[l] * procinst->darken))/3), procinst->brighten);
                                                                                trgb[2] = boostup(((ub[l] + bb[l] + (int)((double)b[l] * procinst->darken))/3), procinst->brighten);

                                                                                // And draw the new line, now antialiased!
                                                                                fa->setRGB(l,(i-procinst->spln)+procinst->thickness,trgb);
                                                                        }

                                                                }

                                                        } // End of the Below Check

                                                        // OK, now do thickness loop and draw the line
                                                        // Set the Pixel line values...

                                                        k = 0;

                                                        while (k<procinst->thickness && i<fa->height) {

                                                                ur = fa->bufLine(LWBUF_RED, i);
                                                                ug = fa->bufLine(LWBUF_GREEN, i);
                                                                ub = fa->bufLine(LWBUF_BLUE, i);

                                                                for (l = 0; l<fa->width; l++) {
                                                                        rgb[0] = boostup((int)((double)ur[l] * procinst->darken), procinst->brighten);
                                                                        rgb[1] = boostup((int)((double)ug[l] * procinst->darken), procinst->brighten);
                                                                        rgb[2] = boostup((int)((double)ub[l] * procinst->darken), procinst->brighten);
                                                                        fa->setRGB(l,i,rgb);
                                                                }

                                                                i++;
                                                                k++;

                                                        } // End of the Thickness loop

                                                        i--;    // Back one line otherwise get black lines if no antialiasing...

                                                }

                                else

                                        {

                                        // The line was to be drawn normally
                                        // Grab pointers to the RGBA values.

                                        r = fa->bufLine(LWBUF_RED, i);
                                        g = fa->bufLine(LWBUF_GREEN, i);
                                        b = fa->bufLine(LWBUF_BLUE, i);

                                        j = 0;

                                        while (j<fa->width) {

                                                rgb[0] = boostup(r[j], procinst->brighten);
                                                rgb[1] = boostup(g[j], procinst->brighten);
                                                rgb[2] = boostup(b[j], procinst->brighten);
                                                fa->setRGB(j,i,rgb);
                                                j++;

                                                } // End of the Loop across the image

                                } // End of normal line

                                i++;    // Increment height counter

                } // End of the Loop down the Image

                return;
        }



// Interface Function

XCALL_(int) theinterface (
                                long            version,
                                GlobalFunc      *global,
                                LWInstance      inst,
                                void            *serverdata
                                )
{
                // First step - Define some variables for our panel...
                // A pointer to panel functions, and an ID are needed, etc...
                // (Read the SDK Doc's for more info (gui_supp.zip))

                LWPanelFuncs            *panl;
                LWPanelID               panID;
                LWControl               *ctrl;
                MessageFuncs            *message;

                // Create the Instance to store our data into
                VIDIMAGE *intinst = (VIDIMAGE *)(inst);

                // OK, Now we need to set up our Services.. This also checks to see if the
                // LWPanels.p plugin is installed...

                // Prepare the message functions
                message = (MessageFuncs *)(global("Info Messages", GFUSE_TRANSIENT));
                if (!message) return AFUNC_BADGLOBAL;

                // Get the Panel functions
                panl = (LWPanelFuncs *)(global(PANEL_SERVICES_NAME,GFUSE_TRANSIENT));
                if (!panl) {
                                message->error("Unable to activate global"PANEL_SERVICES_NAME,"   Please add plugin lwpanels.p");
                                return AFUNC_BADGLOBAL;
                                }

                // Ok, all the functions have been activated (or aborted!) so the next
                // Step is to create our user interface...

                // Create the panel in memory
                panID = panl->create("T42 Video Lines Filter", NULL);

                // Create our Selection Array
//                const char * fxtype[][7] = {"(None)","Above","Below","Both"};

                // We need controls for the Startline, Steps, Soft effect and Line colour...
                INT_CTL(panl, panID, "Line Thickness:");
                INT_CTL(panl, panID, "Line Steps:");
                HCHOICE_CTL(panl,panID, "Soften Line:",fxtype[0]);
                FLOAT_CTL(panl, panID, "Darkness of lines %:");
                FLOAT_CTL(panl, panID, "Boost brightness  %:");

                // Now set the values
                ctrl = panl->nextControl(panID, NULL);
                SET_INT(ctrl, intinst->thickness);
                ctrl = panl->nextControl(panID, ctrl);
                SET_INT(ctrl, intinst->spln);
                ctrl = panl->nextControl(panID, ctrl);
                SET_INT(ctrl, intinst->dirn);
                ctrl = panl->nextControl(panID, ctrl);
                SET_FLOAT(ctrl, 100 - ((intinst->darken)*100));
                ctrl = panl->nextControl(panID, ctrl);
                SET_FLOAT(ctrl, (intinst->brighten)*100);


                // OK, now we need to display the panel, and wait for Mouse reponse
                // The returned values are 1 for OK or 0 for cancel...

                panlstate = panl->open(panID, (PANF_BLOCKING | PANF_CANCEL));

                // Here we would do our checks on the state of panlstate...
                // If 1 then grab our data into our data by using:

                if (panlstate) {

                        // Get the Thickness of the lines
                        ctrl = panl->nextControl(panID, NULL);
                        GET_INT(ctrl, intinst->thickness);

                        // Get the step rate of the lines
                        ctrl = panl->nextControl(panID, ctrl);
                        GET_INT(ctrl, intinst->spln);

                        // Get the Softfx Direction
                        ctrl = panl->nextControl(panID, ctrl);
                        GET_INT(ctrl, intinst->dirn);

                        // Get the Darkness Factor
                        ctrl = panl->nextControl(panID, ctrl);
                        GET_FLOAT(ctrl, intinst->darken);

                        if (intinst->darken > 100){

                                intinst->darken = 100;
                                message->info("Darken Amount can only","be a maximum of 100%");
                        }

                        intinst->darken = (100 - intinst->darken) / 100;

                        // Get the Brightness Factor
                        ctrl = panl->nextControl(panID, ctrl);
                        GET_FLOAT(ctrl, intinst->brighten);

                        intinst->brighten = intinst->brighten / 100;

                        if (intinst->thickness >= intinst->spln) {

                                intinst->thickness = intinst->spln /2;
                                message->warning("Thickness exceeded or equalled step size.","Thickness Value adjusted to half step size.");

                        }

                }

                // Close the panel now that we are completed

                panl->destroy(panID);

                return AFUNC_OK;

}

// That's it - Our plugin is complete!

